/* cacher.c - Cache du texte dans une image
 * ____________________________________________________________________
 * En utilisant les principes de la steganographie, on cache du texte
 * dans une image. On stocke un nombre de bits (note n )passes en
 * parametre dans les n bits de poids faibles d'une composante d'un
 * pixel. Ainsi, si n=4, on stockera 4 bits de d'un caractere
 * (soit la moitie d'un caractere) dans les 4 bits de poids faibles
 * d'un pixel. Pour pouvoir decoder le fichier, n sera stocke dans
 * la composante rouge du 1er pixel, les 2 autres composantes restant
 * inchangees.
 * ____________________________________________________________________
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "io.h"
#include "progress.h"
#include "effets/cacher.h"
#include "effets/cacher_lib.h"

#define TAILLE_BUFFER 1024

int main(int argc, char** argv)
{
	int leRetour,errflg;
	struct image lImage;

	/* Les options recuperes de la ligne de commande */
	int leCoef;
	char* leFichierSrc;
	char* leFichierDest;
	char* leFichierTexte;

	/* Initialisations */
	errflg=0;
	leCoef=0;
	leFichierSrc=NULL;
	leFichierDest=NULL;
	leFichierTexte=NULL;

	/* Analyse des arguments */
	if(argc!=4)
	{
		errflg++;
	}
	else
	{
		/* Le coefficient */
		leCoef=atoi(argv[1]);
		if(leCoef==0) errflg++;

		/* Allocation des chaines */
		leFichierTexte=(char*)ppm_malloc(sizeof(char)*strlen(argv[2])+1);
		leFichierSrc=(char*)ppm_malloc(sizeof(char)*strlen(argv[3])+1);

		/* Copie des chaines */
		strcpy(leFichierTexte,argv[2]);
		strcpy(leFichierSrc,argv[3]);
	}

	if(errflg)
	{
		fprintf(stderr,"%s [indice] [fic_texte] [source]\n",argv[0]);

		if(leFichierTexte!=NULL)
		{
			ppm_free((void*)leFichierSrc);
			ppm_free((void*)leFichierTexte);
		}
		return ERROR_ARGS;
	}

	/* Recupere l'image */
	leRetour=get_image(&lImage, leFichierSrc);
	if(leRetour==NO_ERROR)
	{
		/* Applique l'effet */
		
		leRetour=cacher(&lImage, leFichierTexte, leCoef);
		if(leRetour==NO_ERROR)
		{
			/* Generation du nom de fichier */
			leFichierDest=get_file_name(argv[3],"_C.ppm");

			/* Enregistre l'image */
			leRetour=create_image(&lImage, leFichierDest);

			/* Desallocation */
			ppm_free((void*)leFichierDest);
		}
	}

	/* Desallocation */
	ppm_free((void*)leFichierSrc);
	ppm_free((void*)leFichierTexte);

	return leRetour;
}

int cacher(struct image* telleImage, const char* telNomFichier, int telCoef)
{
	FILE* leFichier;
	int i, j, cont, rang, leCoef, laTaille, laTailleImage;
	char* leContenuFichier;
	char* leContenuFichierBin;
	char* leTabTmp;
	char* leTabRVB;
	char c;

	/* Initialisation */
	leTabRVB=NULL;
	c=0;

	/* Test du coef passe en parametre, il doit etre egal a 1, 2 ou 4 */
	if(telCoef!=1 && telCoef!=2 && telCoef!=4)
	{
		fprintf(stderr,"Le coefficient doit etre 1, 2 ou 4 !\n");
		return ERROR_ARGS;
	}
	else
	{
		leCoef=8/telCoef;
	}

	if((leFichier=fopen(telNomFichier, "r"))==NULL)
	{
		fprintf(stderr,"Fichier texte inexistant !\n");
		return ERROR_OPEN_FILE;
	}

	/* On peut remarquer que l'on fait deux boucles while (presque)
	 * identiques:
	 * une pour calculer le nombre de caracteres du fichier texte,
	 * l'autre pour mettre le contenu du fichier dans un char*.
	 *
	 * On aurait pu en effet, n'utiliser qu'une seule boucle while
	 * qui aurait fait ceci:
	 *
	 * laTaille=0;
	 * while((c=getc(leFichier))!=EOF)
	 * {
	 *	leContenuFichier=(char*)realloc(leContenuFichier,
						laTaille+sizeof(char));
	 *	leContenuFichier[laTaille]=c;
	 *	laTaille++;
	 * }
	 *
	 * Le probleme de cette boucle, c'est le realloc. En effet, cette
	 * fonction est "lente" a s'executer surtout dans une boucle contenant
	 * beaucoup d'iterations, comme c'est le cas ici. Le but c'est
	 * d'effectuer le moins de realloc possible.
   	 * Le meilleur moyen, c'est de connaitre le nombre de caracteres du
	 * fichier et de ne faire qu'un et un seul malloc.
	 * Nous sommes alors obliges d'effectuer deux boucles while, 1 pour
	 * connaitre cette fameuse taille, l'autre pour copier le fichier dans
	 * un char*.
	 *
   	 * Cette methode pourrait s'averer pire que la succession de realloc
	 * mais il n'en est rien. En effet, la complexite de la seconde methode
	 * (les deux boucles while) est de l'ordre de 2*N, au pire.
	 * Alors que le while avec le realloc a une complexite d'ordre N!,
	 * au pire.
	 * N=leNbDeCaracDuFichier ici et on considere que la complexite de getc
	 * est negligeable.
	 * ==> Il est donc preferable d'effectuer deux boucles while
	 *     plutot qu'une ...
	 */

	/* Calcul du nombre de caracteres du fichier */
	laTaille=0;
	while(getc(leFichier)!=EOF)
	{
		laTaille++;
	}

	leContenuFichier=(char*)ppm_malloc(laTaille*sizeof(char));
	rewind(leFichier);

	/* Recopie du fichier dans un char* */
	i=0;
	while((c=getc(leFichier))!=EOF)
	{
		leContenuFichier[i]=c;
		i++;
	}

	fclose(leFichier);

	/* Verification de la taille de l'image et du nombre de caracteres
	 * possibles d'inserer
	 * -1 car le premier pixel contient le nombre permettant de decoder
	 * le fichier texte
	 */
	laTailleImage=telleImage->saLargeur*telleImage->saHauteur;
	if((3*(laTailleImage-1))/leCoef<laTaille+1)
	{
		ppm_free((void*)leContenuFichier);
		return INVALID_SIZE;
	}

	/* Insertion dans l'image
	 * alloue 8 fois le nb de caracteres du fichier + 1 (\0)
	 */
	leContenuFichierBin=(char*)ppm_malloc(8*(sizeof(char)*laTaille+1)+1);

	/* Construction du tableau binaire par concatenation des
	 * tableaux de 8 bits de chaque caractere
	 */
	cont=0;
	for(i=0; i<laTaille; i++)
	{
		leTabTmp=get_tab_carac(leContenuFichier[i]);
		for(j=0; j<8; j++)
		{
			leContenuFichierBin[cont]=leTabTmp[j];
			cont++;
		}
	}


	/* Ajout de EOF en binaire */
	leTabTmp=get_tab_carac(EOF);
	for(i=0; i<8; i++)
	{
		leContenuFichierBin[cont]=leTabTmp[i];
		cont++;
	}

	/* On stocke tout de meme EOF a la fin du tableau pour que la boucle
	 * puisse s'arreter plus bas dans le for lors de l'insertion de chaque
	 * caractere dans l'image mais ce caractere ne figurera pas dans
	 * l'image de destination.
	 */
	leContenuFichierBin[cont]=EOF;

	ppm_free((void*)leTabTmp);

	/* Stockage du coeficient de codage (1, 2 ou 4) dans la premiere
	 * composante du premier pixel (rouge)
	 * On stocke cette information sur 3 bits, en effet 4 en binaire = 100
	 */
	leTabRVB=rempli_tabRVB(telleImage->sonContenu[0]);

	leTabTmp=get_tab_carac(telCoef);
	leTabRVB[5]=leTabTmp[5];
	leTabRVB[6]=leTabTmp[6];
	leTabRVB[7]=leTabTmp[7];

	telleImage->sonContenu[0]=get_carac_from_tab_bits(leTabRVB, 24);
	ppm_free((void*)leTabTmp);
	printf("Traitement           ");
	fflush(stdout);
	cont=0;
	for(i=1; i<laTailleImage && leContenuFichierBin[cont]!=EOF; i++)
	{
		/*affiche_progress(i, (laTaille*8)/(telCoef*3));*/
		affiche_progress(i,(8/leCoef)*3*laTaille);

		/* Desallocation de leTabRVB avant un nouveau malloc */
		ppm_free((void*)leTabRVB);

		/* Creation d'un tableau de 24 cases (3*8 cases) contenant
		 * les 3 composantes R,V et B en binaire
		 */
		leTabRVB=rempli_tabRVB(telleImage->sonContenu[i]);

		/* leTabRVB contient les 3 composantes en binaire
		 * (a raison de 1 bit par case du tableau)
		 * La seconde etape consiste a parcourir ce tableau en
		 * remplacant les bits correspondant par le caractere
		 * cours du fichier bit par bit.
		 */
		for(j=0; j<3; j++)
		{
			rang=j*8+(8-telCoef);

			while(rang<j*8+8 && leContenuFichierBin[cont]!=EOF)
			{
				leTabRVB[rang]=leContenuFichierBin[cont];
				cont++;
				rang++;
			}
		}

		/* On enregistre le tableau RVB dans l'image a la place de
		 * l'ancienne composante RVB
		 */
		telleImage->sonContenu[i]=get_carac_from_tab_bits(leTabRVB,24);
	}

	ppm_free((void*)leContenuFichier);
	ppm_free((void*)leContenuFichierBin);

	termine_progress("OK");
	return NO_ERROR;
}
