/* decacher.c - Extrait un fichier texte cache dans une image
 *              par le programme cacher
 * ____________________________________________________________________
 * Le nombre de bits par composante est stocke dans les 3 bits de
 * poids faibles de la composante rouge du 1er pixel.
 * ____________________________________________________________________
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "io.h"
#include "progress.h"
#include "effets/cacher_lib.h"
#include "effets/decacher.h"

int main(int argc, char** argv)
{
	struct image lImage;	/* Structure contenant l'image lue	*/
	char* leFichierRetour;	/* Contiendra le texte lu dans l'image	*/
	char* leNomFichierDest;	/* Nom du fichier de destination 	*/
	FILE* leFichierDest;
	int leRetour;

	/* Analyse des arguments */
	if(argc!=2)
	{
		fprintf(stdout,"usage: %s [img_source]\n",argv[0]);
		return ERROR_ARGS;
	}
	
	leRetour=get_image(&lImage, argv[1]);
	if(leRetour==NO_ERROR)
	{
		/* Generation du nom de fichier texte */
		leNomFichierDest=get_file_name(argv[1],".txt");

		/* Applique l'effet */
		leFichierRetour=decacher(&lImage);

		/* Ecriture dans le fichier de destination */
		leFichierDest=fopen(leNomFichierDest, "w");
		if(leFichierDest==NULL)
		{
			leRetour=ERROR_OPEN_FILE;
		}
		else
		{
			fprintf(leFichierDest, "%s", leFichierRetour);
			fclose(leFichierDest);
		}

		/* Desallocation */
		ppm_free((void*)leFichierRetour);
		ppm_free((void*)leNomFichierDest);
		ppm_free((void*)lImage.sonContenu);
	}

	return leRetour;
}

char* decacher(const struct image* telleImage)
{
	int leCoef, i, j, cont, limit, laTailleImage;
	char* leTabRVB;
	char* leFichierBinaire;
	unsigned char leBoolFin;
	char* telFichierRetour;

	/* Initialisation */
	leTabRVB=NULL;
	telFichierRetour=NULL;

	/* Decryptage de la cle sur 3 bits de poids faible dans
	 * la premiere composante R
	 */
	leTabRVB=rempli_tabRVB(telleImage->sonContenu[0]);

	/* On extrait et transforme en entier les 3 derniers bits du tableau */
	leCoef=get_carac_from_tab_bits(leTabRVB+5, 3);

	/* On parcours toutes l'image et on decode avec leCoef
	 * comme 'cle' jusqu'au EOF
	 */
	i=1;
	cont=0;

	/* Il y a eu un malloc sur leTabRVB au dessus et le contenu
	 * n'est plus usite a ce stade
	 */
	ppm_free((void*)leTabRVB);

	printf("Lecture              "); fflush(stdout);
	laTailleImage=telleImage->saLargeur*telleImage->saHauteur;

	/* Premiere partie: Extraction des donnees */
	/* Allocation */
	leFichierBinaire=(char*)ppm_malloc((laTailleImage-1)*3*
						leCoef*sizeof(char));
	while(i<laTailleImage)
	{
		affiche_progress(i, laTailleImage);
		leTabRVB=rempli_tabRVB(telleImage->sonContenu[i]);

		/* On recupere 'leCoef' derniers bits et on les stocke dans
		 * un autre tableau destine a accueillir tous les bits de tout
		 * le fichier image et ainsi reconstituer le fichier originel.
		 */

		/* Recopie des derniers bits de leTabBits a la suite
		 * du tableau leFichierBinaire
		 */
		for(j=8-leCoef; j<8; j++)
		{
			leFichierBinaire[cont]=leTabRVB[j];
			leFichierBinaire[cont+leCoef]=leTabRVB[8+j];
			leFichierBinaire[cont+2*leCoef]=leTabRVB[16+j];

			cont++;
		}

		/* Desallocation de leTabRVB avant un nouveau malloc
		 * execute par la fonction rempli_tabRVB
		 */
		ppm_free((void*)leTabRVB);
		
		cont+=2*leCoef;
		i++;
	}
	termine_progress("OK");
	
	/* Seconde partie: Creation du fichier texte qui est
	 *		   en fait telFichierRetour
	 */
	limit=cont/8;
	cont=0;
	leBoolFin=0;

	printf("Decodage             "); fflush(stdout);
	
	for(i=0; i<limit; i++)
	{
		affiche_progress(i, limit);
		telFichierRetour=(char*)ppm_realloc(telFichierRetour,
							(i+1)*sizeof(char));
		
		
		telFichierRetour[i]=(char)
			get_carac_from_tab_bits(i*8+leFichierBinaire,8);

		if(telFichierRetour[i]==EOF)
		{
			/* Si on rencontre le caractere EOF, on sort de la
			 * boucle. En effet, on est a la fin du fichier.
			 */
			telFichierRetour[i]='\0';
			leBoolFin=1;
			break;
		}
	}

	if(!leBoolFin)
	{
		/* Probleme, le caractere EOF n'est pas present
		 * On ajoute tout de meme un caractere \0 a la fin de la
		 * chaine pour creer quand meme correctement le fichier.
		 * En effet, cela est due a la demande de l'utilisateur de
		 * decacher un fichier qui ne contient aucun fichier cache...
		 */

		telFichierRetour=(char*)ppm_realloc(telFichierRetour,
					sizeof(char)*(i+1));
		telFichierRetour[i]='\0';
	}

	termine_progress("OK");

	/* Desallocation de la memoire */
	ppm_free((void*)leFichierBinaire);
	
	return telFichierRetour;
}
