/* fusion.c - Fusionne n image de taille differentes
 * ____________________________________________________________________
 * Le fichier de destination contient la superposition des images
 * passes en parametres. La taille de cette image correspond au max des
 * largeurs et au max des hauteurs. Le nombre de couleurs de cette
 * image correspond au max des max couleurs.
 * Si on fusionne 3 images respectivement decomposees par les programmes
 * quelerouge, quelevert et que le bleu, on obtient l'image originale.
 * ____________________________________________________________________
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "io.h"
#include "progress.h"
#include "effets/fusion.h"

int main(int argc,char** argv)
{
	int i,j,leRetour;
	struct image**	lesImages;	/* Les images a fusionner */
	struct image 	lImageDest;	/* L'image resultante     */
	char* leFichierDest;
	int leMaxCol;
	char leType;

	leRetour=NO_ERROR;

	if(argc<3)
	{
		fprintf(stderr,"usage: %s [type] [image_source_1]",argv[0]);
		fprintf(stderr," ... [image_source_n]\n");
		fprintf(stderr,"Type de fusion :\n");
		fprintf(stderr,"\t-rvb : fusion de couleurs (fusion du sujet)\n");
		fprintf(stderr,"\t-normal : fusion d'images\n");
		return ERROR_ARGS;
	}

	/* Lit les images passees en parametre dans un tableau de structure */
	lesImages=(struct image**)ppm_malloc((argc-2)*sizeof(struct image*));
	for(i=0;i<argc-2;i++)
	{
		/* Allocation d'une nouvelle image */
		lesImages[i]=(struct image*)ppm_malloc(sizeof(struct image));

		/* Lecture de l'image */
		leRetour=get_image(lesImages[i], argv[i+2]);
		if(leRetour!=NO_ERROR)
		{
			/* Desallocation des i-1 premieres images*/
			for(j=0;j<i;j++)
			{
				ppm_free((void*)lesImages[j]->sonContenu);
				ppm_free((void*)lesImages[j]);
			}

			/* Liberation de l'image courante */
			ppm_free((void*)lesImages[i]);

			/* Liberation du tableau */
			ppm_free((void*)lesImages);

			return leRetour;
		}
	}

	if(leRetour==NO_ERROR)
	{
		/* Generation du nom de fichier */
		leFichierDest=get_file_name(argv[2],"_fusion.ppm");

		/* Detection du nombre de couleur max */
		leMaxCol=get_max_col(lesImages,argc-2);

		/* Echantillonage des images qui n'ont pas le bon max col */
		for(i=0;i<argc-2;i++)
		{
			if(lesImages[i]->sonMaxCol!=leMaxCol)
			{
				change_max_color(lesImages[i],leMaxCol);
			}
		}

		/* Detection de la hauteur max et de la largeur max */
		lesImages=canvas_max(lesImages,argc-2);

		/* Type de fusion */
		leType = strcmp(argv[1],"-normal")==0 ? FUS_NORM : FUS_RVB;

		/* Fusion des images */
		fusion(0, argc-2, lesImages, &lImageDest, leType);

		/* Ecriture de l'image fusionnee */
		leRetour=create_image(&lImageDest, leFichierDest);

		/* Desallocation */
		for(i=0;i<argc-2;i++)
		{
			ppm_free((void*)lesImages[i]->sonContenu);
			ppm_free((void*)lesImages[i]);
		}
		ppm_free((void*)lesImages);
		ppm_free((void*)lImageDest.sonContenu);
		ppm_free((void*)leFichierDest);
	}

	return leRetour;
}

void fusion(int telDeb, int telNbImages, struct image** tellesImages,
	    struct image* telleImageDest, char type)
{
	int i, img1, img2,red,green,blue, laTaille;

	/* Au 1er appel, la derniere image du tableau est copiee dans
	 * l'image de destination
	 */
	if(telDeb==0)
	{
		printf("Fusion des images    "); fflush(stdout);
		copie_image(telleImageDest, tellesImages[telNbImages-1]);
	}

	/* Tant qu'il reste des images non fusionnees, on relance la
	 * fusion de l'image de destination avec l'image suivante
	 */
	if(telDeb<telNbImages-1)
	{
		fusion(telDeb+1,telNbImages,tellesImages,telleImageDest, type);
	}

	/* Fusion de l'image en cours avec l'image de destination */
	laTaille=tellesImages[telDeb]->saLargeur *
		 tellesImages[telDeb]->saHauteur;

	for(i=0; i<laTaille; i++)
	{
		affiche_progress(i+telDeb*laTaille,telNbImages*laTaille);

		/* Composantes RVB des 2 images */
		img1=tellesImages[telDeb]->sonContenu[i];
		img2=telleImageDest->sonContenu[i];

		/* Rouge */
		red=valeur_fusion(get_red(img1), get_red(img2),
				  tellesImages[telDeb], telleImageDest, type);
				
		/* Vert */
		green=valeur_fusion(get_green(img1), get_green(img2),
				    tellesImages[telDeb],telleImageDest, type);
		
		/* Bleu */
		blue=valeur_fusion(get_blue(img1), get_blue(img2),
				   tellesImages[telDeb],telleImageDest, type);
		
		/* Ecriture du nouveau pixel */
		telleImageDest->sonContenu[i]=get_code_hexa(red,green,blue);
	}
	
	if(telDeb==0)
	{
		termine_progress("OK");
	}
}

int valeur_fusion(int val1,int val2,struct image* img1, struct image* img2, 
							char type)
{
	switch(type)
	{
		case FUS_RVB :
		
			if(val1==0)
			{
				return val2;
			}
	
			if(val2==0)
			{
				return val1;
			}	

			break;

		case FUS_NORM :
		default:
			
			return val1/2+val2/2;
	

	}

	return val1/2+val2/2;
}

/* Agrandit les images au maximum des tailles horizontales et verticales
 * des images du tableau passe en parametres
 */
struct image** canvas_max(struct image** tellesImages,int telNbImage)
{
	int i,maxLargeur,maxHauteur;
	struct image* telleImage;
	maxLargeur=maxHauteur=-1;

	for(i=0;i<telNbImage;i++)
	{
		if(tellesImages[i]->saLargeur>maxLargeur)
		{
			maxLargeur=tellesImages[i]->saLargeur;
		}

		if(tellesImages[i]->saHauteur>maxHauteur)
		{
			maxHauteur=tellesImages[i]->saHauteur;
		}
	}

	for(i=0;i<telNbImage;i++)
	{
		telleImage=(struct image*)ppm_malloc(sizeof(struct image));
		copie_image_agrandie(telleImage,tellesImages[i],
				     maxLargeur,maxHauteur);
		tellesImages[i]=telleImage;
	}

	return tellesImages;
}

/* Retourne le nombre de couleurs maximum du tableau d'images
 * passes en parametre
 */
int get_max_col(struct image** tellesImages,int telNbImage)
{
	int maxColor;
	int i;
	maxColor=-1;

	for(i=0;i<telNbImage;i++)
	{
		if(tellesImages[i]->sonMaxCol>maxColor)
		{
			maxColor=tellesImages[i]->sonMaxCol;
		}
	}

	return maxColor;
}
