/* io.c - Entrees/Sorties de fichiers ppm
 * ____________________________________________________________________
 * Rappel sur le format PPM (extrait du man) :
 *  Un fichier ppm contient :
 *	- Un nombre magique (le type du fichier sur 2 caracteres)
 * 	  sera toujours P3 dans notre cas
 *	- Des separateurs (espaces, tabulations, CR, LF)
 *	- La largeur
 *	- Des separateurs
 *	- La hauteur
 *	- Des separateurs
 *	- Le nombre maximal de couleurs
 *	- Des separateurs
 *	- Les caracteres a partir d'un # jusqu'a la fin d'une ligne,
 *	  avant le nombre maximal de couleurs, sont des commentaires
 *	  et sont ignores
 * _____________________________________________________________________
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "io.h"
#include "error.h"
#include "progress.h"
#define TAILLE_BUFFER 255

/* Caractere courant lu dans le fichier */
int leCaractereCourant;

/* Lit une image au format ppm depuis un fichier
 *
 * param : telImage : 	  structure amenee a contenir l'image lue
 * param : telNomFichier : nom du fichier a lire
 *
 * retour : code d'erreur (ou de succes)
 */
int get_image(struct image* telleImage, const char* telNomFichier)
{
	FILE*	leFichier;		/* Pointeur vers le fichier 	*/
	char	leBuffer[TAILLE_BUFFER];/* Buffer de lecture 		*/
	int	laLargeur;		/* Largeur lue			*/
	int	laHauteur;		/* Hauteur lue			*/
	int	laTailleImage;		/* Nombre de pixels de l'image	*/
	int	leMaxCouleur;		/* Nombre maximal de couleur lu */
	int 	leRouge;		/* Composante rouge d'un pixel  */
	int	leVert;			/* Composante verte d'un pixel  */
	int 	leBleu;			/* Composante bleue d'un pixel  */
	int	i;

	printf("Chargement du fichier");
	fflush(stdout);

	/* Verification de l'existence du fichier a ouvrir */
	leFichier=fopen(telNomFichier,"r");
	if(leFichier==NULL)
	{
		termine_progress("FICHIER INEXISTANT");
		return ERROR_OPEN_FILE;
	}

	/* Recupere le premier caractere du fichier */
	leCaractereCourant=fgetc(leFichier);

	/* Recupere et verifie le format */
	get_a_token(leFichier,leBuffer);
	if(strncmp(leBuffer, FORMAT,2)!=0)
	{
		termine_progress("FORMAT INVALIDE");
		fclose(leFichier);
		return INVALID_FORMAT;
	}

	/* Recupere et verifie les dimensions */
	get_a_token(leFichier,leBuffer);
	laLargeur=isdigit(*leBuffer)?atoi(leBuffer):0;

	get_a_token(leFichier,leBuffer);
	laHauteur=isdigit(*leBuffer)?atoi(leBuffer):0;

	/* Recupere et valide le nombre max de couleurs */
	get_a_token(leFichier,leBuffer);
	leMaxCouleur=isdigit(*leBuffer)?atoi(leBuffer):0;

	/* Sort si les dimensions ou le maximum de couleurs non valides */
	if(laLargeur==0 || laHauteur==0 || leMaxCouleur==0)
	{
		termine_progress("ENTETE INVALIDE");
		fclose(leFichier);
		return INVALID_SIZE;
	}

	/*
	 * Lecture des donnees (Composantes RVB)
	 * ===================================================================
	 *
	 *  Allocation d'une zone de memoire contigues (destinee a
	 *  contenir des entiers) contenant (largeur * hauteur) cases
	 *
	 * => 1 case par composante RVB stocke de la maniere suivante :
	 *	R : ab  |
	 *	V : cd  | en hexadecimal
	 *	B : ef  |
	 *
	 *    Ce qui donne un nombre hexadecimal de la forme : abcdef
	 *
	 * => Ex: 255 0 50
	 *	R = 255 --HEXA--> FF
	 *	V = 0   --HEXA--> 00
	 *	B = 50  --HEXA--> 32
	 *
	 * On stockera donc la valeur 0xFF0032 pour coder cette composante
	 *
	 * CONCLUSION :
	 * ------------
	 * CETTE METHODE PERMET DONC DE NE STOCKER QUE largeur*hauteur
	 * entiers PLUTOT QUE largeur*hauteur*3 entiers.
	 *
	 */

	i=0;
	laTailleImage=laLargeur*laHauteur;
	telleImage->sonContenu=(int*)ppm_malloc(sizeof(int)*laTailleImage);

	while(i<laTailleImage)
	{
		/* Recupere la composante rouge */
		get_a_token(leFichier,leBuffer);
		leRouge=isdigit(*leBuffer)?atoi(leBuffer):-1;

		/* Recupere la composante verte */
		get_a_token(leFichier,leBuffer);
		leVert=isdigit(*leBuffer)?atoi(leBuffer):-1;

		/* Recupere la composante bleu */
		get_a_token(leFichier,leBuffer);
		leBleu=isdigit(*leBuffer)?atoi(leBuffer):-1;

		/* Test si la couleur est bien un nombre */
		if(leRouge==-1 || leVert==-1 || leBleu==-1)
		{
			termine_progress("CORPS INVALIDE");
			ppm_free((void*)telleImage->sonContenu);
			fclose(leFichier);
			return INVALID_DATA;
		}

		/* Chaque couleur doit etre inferieure au max de couleurs */
		if(leRouge>leMaxCouleur || leVert>leMaxCouleur ||
		   leBleu>leMaxCouleur)
		{
			termine_progress("VALEUR HORS LIMITES");
			ppm_free((void*)telleImage->sonContenu);
			fclose(leFichier);
			return INVALID_DATA;
		}

		affiche_progress(i,laTailleImage);

		/* Ecriture du code hexa correspondant a la composante RVB */
		telleImage->sonContenu[i]=get_code_hexa(leRouge,leVert,leBleu);

		i++;
	}

	/* Si trop de composantes alors on sort */
	get_a_token(leFichier,leBuffer);
	if(leBuffer[0]!='\0')
	{
		termine_progress("UEOF");
		ppm_free((void*)telleImage->sonContenu);
		fclose(leFichier);
		return INVALID_DATA;
	}

	/* Si on a bien lu le nombre de composantes prevu */
	if(i==laTailleImage)
	{
		termine_progress("OK");
	}
	else
	{
		termine_progress("EOF");
		ppm_free((void*)telleImage->sonContenu);
		fclose(leFichier);
		return INVALID_DATA;
	}

	/* Ecriture des attributs de l'image */
	telleImage->saLargeur=laLargeur;
	telleImage->saHauteur=laHauteur;
	telleImage->sonMaxCol=leMaxCouleur;

	/* Fermeture du fichier */
	fclose(leFichier);

	return NO_ERROR;
}

/* Recupere un token dans le fichier
 *
 * param : telFichier : Fichier a lire
 * param : telBuffer : Buffer de lecture
 *
 * retour : pas de retour
 */
void get_a_token(FILE* telFichier, char* telBuffer)
{
	int i=0;

	/* Recupere le 1er char qui n'est ni un separateur ni un commentaire */
	leCaractereCourant=delete_whitespaces(telFichier);

	/* Boucle tant que ce n'est ni un separateur ni un commentaire */
	while(!feof(telFichier) && !isspace(leCaractereCourant)
				&& leCaractereCourant!=CHAR_COMMENTAIRE)
	{
		telBuffer[i]=leCaractereCourant;
		leCaractereCourant=fgetc(telFichier);
		i++;
	}

	telBuffer[i]='\0';
}

/* Supprime les espaces (tabs, \n, \r ...)
 * et les commentaires
 *
 * param : telFichier : Fichier a lire
 *
 * retour : le caractere lu
 */
char delete_whitespaces(FILE* telFichier)
{
	/* ETAT =  1 -> On est dans un commentaire
	 * ETAT =  0 -> Lecture de "caracteres normaux" (pas un commentaire)
	 * ETAT = -1 -> On sort (On a lu un token et on arrive a un separateur
	 */

	int etat=0;
	while(!feof(telFichier) && etat!=-1)
	{
		if(etat==0)
		{
			if(leCaractereCourant==CHAR_COMMENTAIRE)
			{
				/* On entre dans un commentaire */
				etat=1;
				leCaractereCourant=fgetc(telFichier);
			}
			else
			{
				if(!isspace(leCaractereCourant)) etat=-1;
				else leCaractereCourant=fgetc(telFichier);
			}
		}
		else
		if(etat==1)
		{
			if(leCaractereCourant=='\n')
			{
				/* C'est la fin d'un commentaire */
				etat=0;
			}
			leCaractereCourant=fgetc(telFichier);
		}
	}

	return leCaractereCourant;
}

/* Ecrit une image au format ppm P3 vers un fichier
 *
 * param : telImage : structure contenant les donnees de l'image
 * param : telNomFichier : nom du fichier a ecrire
 *
 * retour : code d'erreur (ou de succes)
 */
int create_image(struct image* telleImage, const char* telFichier)
{
	FILE* 	leFichier;
	int	laTailleImage;	/* Nombre de pixels de l'image */
	char	leBuffer[TAILLE_BUFFER];
	int 	i;

	printf("Ecriture du fichier  ");
	fflush(stdout);

	/* Ouverture en ecriture du fichier de destination */
	leFichier=fopen(telFichier, "w");
	if(leFichier==NULL)
	{
		return ERROR_WRITE_FILE;
	}

	/* Ecriture du format du fichier */
	if(fprintf(leFichier, "%s\n", FORMAT)<0)
	{
		fclose(leFichier);
		return ERROR_WRITE_FILE;
	}

	/* Ecriture des commentaires */
	strcpy(leBuffer,CHAR_COMMENTAIRE_CHAINE" Projet C Ansi\n");
	strcat(leBuffer,CHAR_COMMENTAIRE_CHAINE" Nicolas Maupu - ");
	strcat(leBuffer,"Matthieu Lomon\n");

	if(fprintf(leFichier,"%s",leBuffer)<0)
	{
		fclose(leFichier);
		return ERROR_WRITE_FILE;
	}

	/* Ecriture des dimensions et du nombre de couleur max de l'image */
	if(fprintf(leFichier,"%d %d\n%d\n", telleImage->saLargeur,
					    telleImage->saHauteur,
					    telleImage->sonMaxCol)<0)
	{
		fclose(leFichier);
		return ERROR_WRITE_FILE;
	}

	/* Ecriture du corps de l'image */
	i=0;
	laTailleImage=telleImage->saLargeur*telleImage->saHauteur;

	while(i<laTailleImage)
	{
		affiche_progress(i,laTailleImage);

		/* Ecriture d'une composante RVB */
		if(fprintf(leFichier,"%d %d %d\n",
			 	get_red(telleImage->sonContenu[i]),
				get_green(telleImage->sonContenu[i]),
				get_blue(telleImage->sonContenu[i]))<0)
		{
			fclose(leFichier);
			return ERROR_WRITE_FILE;
		}

		i++;
	}

	termine_progress("OK");
	fclose(leFichier);
	return NO_ERROR;
}

/* Permet de rajouter un suffixe a une chaine avec le point
 * Si le nom genere correspondant a un fichier deja existant,
 * on ajoute un numero a la fin, et on recommence la verification ...
 *
 * param : telNom : nom du fichier d'origine
 * param : telleChaine : suffixe a ajouter
 * retour: pointeur vers la chaine generee
 */
char* get_file_name(const char* telNom, const char* telleChaine)
{
	char* leNouveauNom;
	int laLongueur;
	int i;
	int lIndPoint;

	lIndPoint=strlen(telNom);
	laLongueur=strlen(telNom);

	/* Recupere l'indice du dernier point dans la chaine */
	for(i=0;i<laLongueur;i++)
	{
		if(telNom[i]=='.')
		{
			lIndPoint=i;
		}
	}

	/* Allocation d'une chaine de la longueur de la chaine jusqu'au point
	 * plus la longueur du suffixe + \0
	 */
	leNouveauNom=(char*)ppm_malloc(sizeof(char)*
				(lIndPoint+strlen(telleChaine)+1));

	/* Coupe le nom juste avant le point */
	for(i=0;i<lIndPoint;i++)
	{
		leNouveauNom[i]=telNom[i];
	}

	/* Ajoute le suffixe telleChaine */
	laLongueur=strlen(telleChaine)+1;
	for(i=0;i<laLongueur;i++)
	{
		leNouveauNom[i+lIndPoint]=telleChaine[i];
	}

	return leNouveauNom;
}
