/* libimage.c - Fonctions de manipulation d'images au format PPM
 * ____________________________________________________________________
 * Ce fichier contient des fonctions utiles a la manipulation de
 * fichiers au format PPM :
 *	- Des fonctions de copies de structures completes, seulement
 *        des entetes, d'echantillonnage ...
 *	- La redefinition de malloc,realloc et free
 * ____________________________________________________________________
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "libimage.h"
#include "error.h"

/* Copier la structure representant une image dans une nouvelle structure */
void copie_image(struct image* telleImageDest,
		 const struct image* telleImageSrc)
{
	int i;
	int laTaille;

	/* Copie des proprietes de l'image */
	copie_image_entete(telleImageDest,telleImageSrc);

	/* Copie du corps de l'image */
	laTaille=telleImageSrc->saLargeur*telleImageSrc->saHauteur;
	for(i=0; i<laTaille; i++)
	{
		telleImageDest->sonContenu[i]=telleImageSrc->sonContenu[i];
	}
}

/* Copier la structure representant une image dans une nouvelle structure
 * dont les pixels ne sont pas initialise 
 */
void copie_image_entete(struct image* telleImageDest,
			const struct image* telleImageSrc)
{
	copie_image_entete_px(telleImageDest,
			      telleImageSrc,
			      telleImageSrc->saLargeur,
			      telleImageSrc->saHauteur);
}

/* Copier la structure representant une image dans une nouvelle structure dont
 * les pixels ne sont pas initialise.
 * Rmq : par rapport a la fonction precedente, on peut allouer un nombre
 *       differents de pixels
 */
void copie_image_entete_px(struct image* telleImageDest,
			   const struct image* telleImageSrc,
			   int telleLargeur, int telleHauteur)
{
	int laTailleImg;
	laTailleImg=telleLargeur*telleHauteur;

	/* Copie des proprietes de l'image */
	telleImageDest->saLargeur=telleLargeur;
	telleImageDest->saHauteur=telleHauteur;
	telleImageDest->sonMaxCol=telleImageSrc->sonMaxCol;

	/* Allocation du contenu */
	telleImageDest->sonContenu=(int*)ppm_malloc(laTailleImg*sizeof(int));
}

/* Copie une image en agrandissant la taille de son canvas */
void copie_image_agrandie(struct image* telleImageDest,
			  const struct image* telleImageSrc,
			  int telleLargeur, int telleHauteur)
{
	int i,j,tmp,curLig;
	int courant;
	int leMaxCol;
	int laTaille;

	laTaille=telleLargeur*telleHauteur; /* Nombre de pixels du canvas */
	leMaxCol=telleImageSrc->sonMaxCol/2;  /* Nombre de couleurs maxi    */

	/* Copie l'entete de telleImage src dans telleImageDest
	 * avec modification de taille
	 */
	copie_image_entete_px(telleImageDest, telleImageSrc,
			      telleLargeur,   telleHauteur);

	/* Copie l'image source collee en haut a gauche.
	 * Le vide est complete par du blanc.
	 */
	courant=0;
	for(i=0;i<telleImageSrc->saHauteur;i++)
	{
		curLig=i*telleImageSrc->saLargeur;
		
		/* Copie de l'image source */
	   	for(j=0;j<telleImageSrc->saLargeur;j++)
		{
			tmp=telleImageSrc->sonContenu[curLig+j];
			telleImageDest->sonContenu[courant]=tmp;
			courant++;
		}
		
		/* On complete la ligne avec du blanc */
		for(;j<telleImageDest->saLargeur;j++)
		{
			telleImageDest->sonContenu[courant]=
				get_code_hexa(leMaxCol, leMaxCol, leMaxCol);
			courant++;
		}
	}

	for(;i<telleImageDest->saHauteur;i++)
	{
		for(j=0;j<telleImageDest->saLargeur;j++)
		{
			telleImageDest->sonContenu[courant]=
				get_code_hexa(leMaxCol, leMaxCol, leMaxCol);
			courant++;
		}        	
	}


}

/* Permet de changer le nombre de couleurs d'une image */
void change_max_color(struct image* telleImage, int telleNbCouleur)
{
	int i;
	int laTaille; 
	int telExMaxCol;

	telExMaxCol=telleImage->sonMaxCol;    /* L'ancien max couleur */
	telleImage->sonMaxCol=telleNbCouleur; /* Nouveau max couleur  */

	laTaille=telleImage->saLargeur*telleImage->saHauteur;
	for(i=0; i<laTaille; i++)
	{
		telleImage->sonContenu[i] *= telleNbCouleur/telExMaxCol;
	}
}

/* Redefinition du malloc */
void* ppm_malloc(size_t size)
{
	void* lePointeur;
	lePointeur=malloc(size);
	
	#ifdef DEBUG
	printf("\nMalloc de taille %d ",size);
	#endif
	
	if(lePointeur==NULL)
	{
		#ifdef DEBUG
		printf("\tErreur de malloc\n");
		#endif
		
		exit(ERROR_MALLOC);	
	}
	else
	{
		#ifdef DEBUG
		printf("\n");
		#endif
		
		return lePointeur;	
	}
}

/* Redefinition du malloc */
void* ppm_realloc(void* ptr,size_t size)
{
	void* lePointeur;
	lePointeur=realloc(ptr,size);
	
	#ifdef DEBUG
	printf("\nRealloc de taille %d ",size);
	#endif
	
	if(lePointeur==NULL)
	{
		#ifdef DEBUG
		printf("\tErreur de malloc\n");
		#endif
		
		exit(ERROR_MALLOC);	
	}
	else
	{
		#ifdef DEBUG
		printf("\n");
		#endif
		
		return lePointeur;	
	}
}

/* Redefinition de free */
void ppm_free(void* ptr)
{
	#ifdef DEBUG
	printf("\nFree\n");	
	#endif

	free(ptr);
}
